# /usr/bin/env python

""" NLP Modelling Classes """

import os
import joblib
from shutil import rmtree
from tempfile import mkdtemp
from sklearn import set_config
from sklearn.preprocessing import FunctionTransformer
from sklearn.pipeline import Pipeline, FeatureUnion
from .vectorizers import Vectorizer
from .estimation import Classifier, BinaryThresholder
from .feature_transform import FeatureScaler, DimensionalityReduction
from .crossval import CrossValidator, count_hyperparameters


def column_selector(data, column_name, reshape=False):
    """
    Column selector function for feature concatenation
    Args:
        data: pandas.DataFrame
        column_name: string
    Returns:
        data: pandas.DataFrame
    """
    if reshape:
        return data.loc[:, column_name].values.reshape(-1, 1)
    else:
        return data.loc[:, column_name].values


def construct_text_pipelines(
    q_col, a_col, env="scikit-learn", model="TfidfVectorizer", **kwargs
):
    """
    Construct Text-Pipeline
    Args:
        column_names:
                list of column names
        kwargs:
                dictionary of keyword arguments for vectorization
        env:
                string, type of model to use
        name:
                string, name of model to use
    Returns:
        text_pipeline:
                sklearn.pipeline.Pipeline
    """
    if q_col is not None:
        q_pipe = Pipeline(
            [
                (
                    "selector",
                    FunctionTransformer(
                        column_selector,
                        kw_args=dict(column_name=q_col),
                        validate=False,
                    ),
                ),
                (
                    "vectorizer",
                    Vectorizer(model, env=env, **kwargs),
                ),
            ]
        )
    if a_col is not None:
        a_pipe = Pipeline(
            [
                (
                    "selector",
                    FunctionTransformer(
                        column_selector,
                        kw_args=dict(column_name=a_col),
                        validate=False,
                    ),
                ),
                (
                    "vectorizer",
                    Vectorizer(model, env=env, **kwargs),
                ),
            ]
        )
    if q_col is not None and a_col is not None:
        text_pipeline = FeatureUnion([("question", q_pipe), ("answer", a_pipe)])
    elif q_col is not None:
        text_pipeline = FeatureUnion([("question", q_pipe)])
    elif a_col is not None:
        text_pipeline = FeatureUnion([("answer", a_pipe)])
    else:
        text_pipeline = None
    return text_pipeline


class Model:
    """
    Main NLP Class for Model Training and Prediction.
    """

    def __init__(self, enable_caching=False, diagram=True):
        """
        Initialize Model
        Args:
            enable_caching (bool): Enable caching of the model
            diagram (bool): Create model diagram
        """
        if enable_caching is True:
            self.cache_dir = mkdtemp(prefix="iqual_")
        else:
            self.cache_dir = None
        self.diagram = diagram
        self.text_vector_pipe = None
        self.transformation_steps = []
        self.estimation_steps = []
        self.post_estimation_steps = []

    def add_text_features(
        self, q_col, a_col, env="scikit-learn", model="TfidfVectorizer", **kwargs
    ):
        """
        Configure text features

        Args:
            q_col (str): Name of the question column
            a_col (str): Name of the answer column
            env (str): type of the vectorizer model to be used for text features
            name (str): name of the vectorizer model to be used for text features
            kwargs: keyword arguments for vectorization
        """

        self.text_vector_pipe = construct_text_pipelines(
            q_col, a_col, env=env, model=model, **kwargs
        )
        self.input_steps = [("Input", self.text_vector_pipe)]
        self.vect_kwargs = kwargs
        self.env = env
        self.model = model

    def add_feature_scaler(self, name="StandardScaler", **kwargs):
        """
        Scale the features
        Args:
            name (str):
                Name of the feature scaling method
            **kwargs:
                Keyword arguments for the scaler method
        """
        self.transformation_steps.append(
            ("FeatureTransformation", FeatureScaler(name=name, **kwargs))
        )

    def add_dimensionality_reducer(self, name="PCA", **kwargs):
        """
        Reduce the features
        Args:
            name (str): Name of the reducer method
            **kwargs: Keyword arguments for the reducer method
        """
        self.transformation_steps.append(
            ("FeatureTransformation", DimensionalityReduction(name=name, **kwargs))
        )

    def add_feature_transformer(
        self, name="StandardScaler", transformation="DimensionalityReduction", **kwargs
    ):
        """ """
        if transformation == "DimensionalityReduction":
            self.add_dimensionality_reducer(name=name, **kwargs)
        elif transformation == "FeatureScaler":
            self.add_feature_scaler(name=name, **kwargs)
        else:
            raise ValueError(
                "Invalid transformation method, please choose from 'DimensionalityReduction' or 'FeatureScaler'"
            )

    def add_classifier(self, name="LogisticRegression", **kwargs):
        """
        Add a classifier to the pipeline
        Args:
            name (str): Name of the classifier
            **kwargs: Keyword arguments for the classifier

        """
        self.estimators = name
        self.kwargs = kwargs
        self.estimation_steps.append(("Classifier", Classifier(name, **kwargs)))

    def add_threshold(self, **kwargs):
        """
        Add a custom threshold to the pipeline
        """
        self.post_estimation_steps.append(("Threshold", BinaryThresholder(**kwargs)))

    def compile(self):
        """
        Compile the model
        """
        ### List of pipeline Steps
        self.steps = [
            *self.input_steps,  # Input (Text Columns)
            *self.transformation_steps,  # Post Vectorization Steps, Feature Scaling
            *self.estimation_steps,  # Estimation Steps
            *self.post_estimation_steps,  # Post Estimation Steps, Custom Threshold
        ]

        #### Create the pipeline
        self.model = Pipeline(self.steps, memory=self.cache_dir)

        if self.diagram is True:
            set_config(display="diagram")
            return self.model

    def cross_validate_fit(
        self,
        X,
        Y,
        cv_method="GridSearchCV",
        search_parameters={},
        scoring="f1",
        cv_splits=5,
        refit=False,
        **kwargs,
    ):
        """
        Cross validate the model and fit it
        Args:
            X (pd.DataFrame): Dataframe of features
            Y (pd.Series): Series of labels
            cv_method (str): Cross validation method
            search_parameters (dict): Search parameters for GridSearchCV
            scoring (str): Scoring method for GridSearchCV
            num_folds (int): Number of folds for GridSearchCV
            n_iter (int): Number of iterations for RandomizedSearchCV
            refit (bool): Whether to refit the model
            **kwargs: Keyword arguments for the model
        """
        print(
            f".......{count_hyperparameters(search_parameters)} hyperparameters configurations possible.....\r",
            end="",
        )

        self.cv = CrossValidator(
            model_pipe=self.model,
            cv_method=cv_method,
            search_parameters=search_parameters,
            scoring=scoring,
            cv=cv_splits,
            refit=refit,
            **kwargs,
        )
        self.cv.fit(X, Y)
        self.set_params(**self.cv.get_best_params())
        self.model.fit(X, Y)
        return self.cv.get_cv_scores()

    def get_params(self):
        """
        Get the parameters of the model
        """
        params = self.model.get_params()
        return params.copy()

    def set_params(self, **kwargs):
        """
        Set the parameters of the model

        """
        self.model.set_params(**kwargs)

    def fit(self, X, y=None, *args, **kwargs):
        """
        Fit the model
        """
        self.model.fit(X, y, *args, **kwargs)
        return self

    def predict_proba(self, X):
        """
        Predict the probabilities of the labels for the positive class
        """
        if hasattr(self.model, "predict_proba"):
            probas = self.model.predict_proba(X)
            if probas.ndim == 2:
                return probas[:, 1]
            elif probas.ndim == 1:
                return probas
            else:
                raise ValueError("The predicted probabilities have an unexpected shape")
        else:
            return self.model.predict(X)

    def predict(self, X):
        """
        Predict the labels  of the samples in X
        """
        return self.model.predict(X)

    def score(self, x, y_true, scoring_function=None):
        """
        Score the model

        Args:
            x (pandas.DataFrame): The data to score
            y_true (pandas.Series): The true labels
            scoring_function (function): The scoring function to use

        """
        y_pred = self.model.predict(x)
        return scoring_function(y_true, y_pred)

    def load_model(self, model_dir, filename):
        """
        Load the model from a pickle file
        Args:
            model_dir (str): Directory where the model is saved
            filename (str): Name of the model file
        """
        model_fp = os.path.join(model_dir, filename + ".pkl")
        self.model = joblib.load(model_fp)

    def save_model(self, model_dir, filename):
        """
        Save the model to disk as a pickle file
        Args:
            model_dir (str): Directory to save the model
            filename (str): Name of the file to save the model
        """
        model_fp = os.path.join(model_dir, filename + ".pkl")
        os.makedirs(model_dir, exist_ok=True)
        joblib.dump(self.model, model_fp)
        print(f"Model saved to {model_fp}")

    def clear_cache(self):
        """
        Clear the cache directory
        """
        if self.cache_dir is not None:
            try:
                rmtree(self.cache_dir)
                print("Cache cleared")
            except FileNotFoundError:
                print("Cache already cleared")
        else:
            print("Cache directory not set")



# /usr/bin/env python
"""
This module contains the vectorizers used to generate embeddings for text data.
Options include:
    - `sentence-transformer` 
    - `spacy`
    - `scikit-learn`
    - `saved-dictionary`
"""
import spacy
import scipy.sparse
import numpy as np
import sklearn.feature_extraction.text
from sentence_transformers import SentenceTransformer
from sklearn.base import BaseEstimator, TransformerMixin
from .utils import save_pickle_data, load_pickle_data


def get_vectorizer_options():
    """Return a list of vectorizer options"""
    return ["scikit-learn", "sentence-transformers", "saved-dictionary", "spacy"]


def load_vectorizer(model, env, **kwargs):
    """
    Load a vectorizer using the `env` and `model` parameters.

    Args:
    """
    assert env in get_vectorizer_options(), f"{env} is not a valid vectorizer option"
    if env == "scikit-learn":
        return ScikitLearnVectorizer(model, **kwargs)
    elif env == "sentence-transformers":
        return SentenceTransformerVectorizer(model, **kwargs)
    elif env == "saved-dictionary":
        return DictionaryVectorizer(model, **kwargs)
    elif env == "spacy":
        return SpacyVectorizer(model, **kwargs)
    else:
        raise ValueError("{} is not a valid env".format(env))


class DictionaryVectorizer(BaseEstimator, TransformerMixin):
    """
    Replace text with vectors using a saved pickle dictionary
    """

    def __init__(self, model="", **kwargs):
        self.model = model

    def load_embeddings(self):
        embedding_dictionary = load_pickle_data(self.model)
        return embedding_dictionary

    def fit(self, *fit_args, **fit_kwargs):
        """ """
        return self

    def transform(self, X, **transform_kwargs):
        """ """
        embeddings = self.load_embeddings()
        return scipy.sparse.csr_matrix(
            np.vstack([embeddings[x] for x in X]), dtype=np.float64
        )

    def get_params(self, **kwargs):
        return {"model": self.model, "env": "saved-dictionary"}

    def set_params(self, **param_args):
        self.model = param_args.get("model", self.model)


class SentenceTransformerVectorizer(BaseEstimator, TransformerMixin):

    """Generate embeddings using the `sentence-transformer` package"""

    def __init__(self, model="average_word_embeddings_glove.6B.300d", **kwargs):
        self.model = model
        self.vectorizer = SentenceTransformer(self.model, **kwargs)

    def fit(self, *fit_args, **fit_kwargs):
        return self

    def transform(self, X, **transform_params):
        """
        Generates embeddings for sentences using the aforementioned initalized `SentenceBert` model
        Args:
             X                   : list of sentences to encode                  (List[str])
             convert_to_numpy    : whether to convert output to numpy array     (Bool)
             normalize_embeddings : whether to normalize embeddings              (Bool)
        Returns:
            Encoded embeddings of sentences

        """
        return self.vectorizer.encode(X, **transform_params)

    def get_params(self, **kwargs):
        return {"model": self.model, "env": "sentence-transformers"}

    def set_params(self, **param_args):
        self.model = param_args.get("model", self.model)
        self.vectorizer = SentenceTransformer(self.model)


class SpacyVectorizer(BaseEstimator, TransformerMixin):
    """Generate vectors using the `spacy` package"""

    def __init__(self, model="en_core_web_sm", **kwargs):
        self.model = model
        self.vectorizer = spacy.load(self.model, **kwargs)

    def fit(self, *fit_args, **fit_kwargs):
        return self

    def transform(self, X, **transform_params):
        return [x.vector for x in self.vectorizer.pipe(X, **transform_params)]

    def get_params(self, **kwargs):
        return {"model": self.model, "env": "spacy"}

    def set_params(self, **param_args):
        self.model = param_args.get("model", self.model)
        self.vectorizer = spacy.load(self.model)


class ScikitLearnVectorizer(BaseEstimator, TransformerMixin):
    """Generate vectors using the `scikit-learn` text-feature extraction modules"""

    def __init__(self, model, **kwargs):
        self.model = model
        self.vectorizer = getattr(sklearn.feature_extraction.text, self.model)()
        self.vectorizer.set_params(
            **{k: v for k, v in kwargs.items() if k not in ["model", "env"]}
        )

    def fit(self, *fit_args, **fit_kwargs):
        """ """
        self.vectorizer.fit(*fit_args, **fit_kwargs)
        return self

    def transform(self, X, **kwargs):
        """ """
        vect = self.vectorizer.transform(X, **kwargs)
        return vect

    def get_params(self, **kwargs):
        params = self.vectorizer.get_params(**kwargs)
        params["model"] = self.model
        params["env"] = "scikit-learn"
        return params

    def set_params(self, **kwargs):
        self.model = kwargs.get("model", "CountVectorizer")
        self.vectorizer = getattr(sklearn.feature_extraction.text, self.model)()
        self.vectorizer.set_params(
            **{k: v for k, v in kwargs.items() if k not in ["model", "env"]}
        )


class Vectorizer(BaseEstimator, TransformerMixin):
    """
    Base class for vectorizers
    """

    def __init__(self, model, env, **kwargs):
        self.model = model
        self.env = env
        self.vectorizer = load_vectorizer(model, env, **kwargs)

    def fit(self, *fit_args, **fit_kwargs):
        """ """
        self.vectorizer.fit(*fit_args, **fit_kwargs)
        return self

    def transform(self, X, **kwargs):
        """ """
        vect = self.vectorizer.transform(X, **kwargs)
        return vect

    def get_params(self, **kwargs):
        params = self.vectorizer.get_params(**kwargs)
        params["model"] = self.model
        params["env"] = self.env
        return params

    def set_params(self, **kwargs):
        self.model = kwargs.get("model", "CountVectorizer")
        self.env = kwargs.get("env", "scikit-learn")
        self.vectorizer = load_vectorizer(
            str(self.model),
            str(self.env),
            **{k: v for k, v in kwargs.items() if k not in ["model", "env"]},
        )



import numpy as np
import pandas as pd
import statsmodels.formula.api as smf

class Bias:
    # TODO: Add support for statsmodels - parent class

    models = smf.__all__ # List of statsapiformula models

    def __init__(
            self,
            enhanced_dataframe,
            annotation_vars=[],
            categorical_regressors=[],
            numerical_regressors=[],
            model='ols',
            act_suffix='_act',
            pred_suffix='_pred',
            error_suffix='_error',
            ):
        """
        Bias test

        Args:
            enhanced_dataframe (pandas dataframe): enhanced dataframe with actual and predicted values
            annotation_vars (list): list of annotation variables
            categorical_regressors (list): list of categorical regressors
            numerical_regressors (list): list of continuous or binary regressors
            model (str): statsmodels model to use. Default is 'ols'
            act_suffix (str): suffix for actual values. This suffix must be present in the enhanced dataframe. Default is '_act'
            pred_suffix (str): suffix for predicted values. This suffix must be present in the enhanced dataframe. Default is '_pred'
            error_suffix (str): suffix for error values. This suffix will be added to the enhanced dataframe. Default is '_error'

        Returns:
            bias_test (Bias): Bias test object
        """
        
        if model not in self.models:
            raise ValueError(f"Model must be one of {', '.join(self.models)}")
        self.model     = model
        self.model_obj = getattr(smf,model)
        self.data      = enhanced_dataframe.copy()
        self.annotation_vars = annotation_vars
        self.categorical_regressors = categorical_regressors
        self.numerical_regressors  = numerical_regressors
        self.model_fits   = {}
        self.act_suffix   = act_suffix
        self.pred_suffix  = pred_suffix
        self.error_suffix = error_suffix
        self._calc_error()

    def _calc_error(self):
        """
        Calculate the error between the actual and predicted values and add it to the dataframe        
        """
        for var in self.annotation_vars:
            self.data[f"{var}{self.error_suffix}"] = self.data[f"{var}{self.act_suffix}"] - self.data[f"{var}{self.pred_suffix}"]
    
    def _format_formula(self,annotation_var):
        """
        Format the formula for the model

        Bias test formula:
            annotation_error ~ C(categorical_regressors) + numerical_regressors

        Args:
            annotation_var (str): annotation variable
        Returns:
            formula (str): statsmodels formula string
        """
        formula = f"{annotation_var}{self.error_suffix} ~ " # Dependent variable
        formula += " + ".join([f"C({var})" for var in self.categorical_regressors]) # Categorical regressors
        formula += " + " + " + ".join([f"{var}" for var in self.numerical_regressors]) # Continuous regressors
        return formula
    
    def fit(self,formula):
        """
        Fit a statsmodels model to the data using the formula
        Args:
            formula (str): statsmodels formula string to fit the model
        Returns:
            model_fit (statsmodels model): fitted model
        """
        model_fit = self.model_obj(formula,data=self.data).fit()
        return model_fit
    
    def fit_all(self):
        """
        Fit all the models.
        All fitted models are stored in self.model_fits
        """
        for var in self.annotation_vars:
            formula = self._format_formula(var)
            self.model_fits[var] = self.fit(formula)
        return self
    def get_model_fits(self):
        """
        Get all the model fits
        Returns:
            model_fits (dict): dictionary of model fits
        """
        return self.model_fits
    
    def get_model_fit(self,annotation_var):
        """
        Get the model fit for a particular annotation variable
        Args:
            annotation_var (str): annotation variable
        Returns:
            model_fit (statsmodels model): fitted model
        """
        return self.model_fits[annotation_var]
    
    def get_model_fit_summary(self,annotation_var):
        """
        Get the model fit summary for a particular annotation variable
        Args:
            annotation_var (str): annotation variable
        Returns:
            model_fit_summary (str): model fit summary
        """
        return self.model_fits[annotation_var].summary()
    
    def get_model_fit_summary_all(self):
        """
        Get the model fit summary for all annotation variables
        Returns:
            model_fit_summary_all (str): model fit summary for all annotation variables
        """
        model_fit_summary_all = ""
        for var in self.annotation_vars:
            model_fit_summary_all += f"{var}: {self.model_fits[var].summary()}"
        return model_fit_summary_all
            
class Interpretability:
    
    models = smf.__all__ # List of statsapiformula models

    def __init__(
            self,
            dataframe,
            annotation_vars=[],
            categorical_regressors=[],
            numerical_regressors=[],
            model='ols',
            ):
        """
        Interpretability class

        Args:
            dataframe (pandas dataframe): dataframe containing the data
            annotation_vars (list): list of annotation variables
            categorical_regressors (list): list of categorical regressors
            numerical_regressors (list): list of continuous or binary regressors
            model (str): statsmodels model to use. Default is 'ols'

        Returns:
            Interpretability object
        """
        # TODO: Add support for statsmodels - parent class
        if model not in self.models:
            raise ValueError(f"Model must be one of {','.join(self.models)}")

        self.dataset = dataframe.copy()
        self.annotation_vars = annotation_vars
        self.categorical_regressors = categorical_regressors
        self.numerical_regressors  = numerical_regressors
        self.model        = model
        self.model_obj    = getattr(smf,model)
        self.model_fits   = {}
        self.model_params = {}

    def _format_formula(self,annotation_var):
        """
        Format the formula for the model.

        Interpretability test formula:
            annotation_var ~ C(categorical_regressors) + numerical_regressors

        Args:
            annotation_var (str): annotation variable
        Returns:
            formula (str): statsmodels formula string
        """

        formula = f"{annotation_var} ~ "
        formula += " + ".join([f"C({var})" for var in self.categorical_regressors])
        formula += " + " + " + ".join([f"{var}" for var in self.numerical_regressors])
        return formula
    
    def fit(self,formula):
        """
        Fit a statsmodels model to the data using the formula.

        Args:
            formula (str): statsmodels formula string to fit the model
        Returns:
            model_fit (statsmodels model): fitted model
        """
        model_fit = self.model_obj(formula,data=self.dataset).fit()
        return model_fit
    
    def fit_all(self):
        """
        Fit all the models. All fitted models are stored in self.model_fits
        """
        for var in self.annotation_vars:
            formula = self._format_formula(var)
            fitted_model = self.fit(formula)
            self.model_fits[var] = fitted_model
            self.model_params[var] = dict(fitted_model.params)            
        return self
    def get_model_fit(self,annotation_var):
        """
        Get the model fit for a particular annotation variable
        Args:
            annotation_var (str): annotation variable
        Returns:
            model_fit (statsmodels model): fitted model
        """
        return self.model_fits[annotation_var]
    
    def get_model_fit_summary(self,annotation_var):
        """
        Get the model fit summary for a particular annotation variable
        Args:
            annotation_var (str): annotation variable
        Returns:
            model_fit_summary (str): model fit summary
        """
        return self.model_fits[annotation_var].summary()
    
    def get_model_params(self,annotation_var):
        """
        Get the model parameters for a particular annotation variable
        Args:
            annotation_var (str): annotation variable
        Returns:
            model_params (dict): model parameters
        """
        return self.model_params[annotation_var]

    def get_fstats(self,annotation_vars=[]):
        """
        Get the F statistics for the annotation variables
        Args:
            annotation_vars (list): list of annotation variables

        Returns:
            fstats (dict): dictionary of F statistics
        """
        fstats = {}
        for var in annotation_vars:
            fstats[var] = self.model_fits[var].fvalue
        return fstats
    
    def get_pvals(self,annotation_vars=[]):
        """
        Get the p-values for the annotation variables.

        Args:
            annotation_vars (list): list of annotation variables
        Returns:
            pvals (dict): dictionary of p-values
        """
        pvals = {}
        for var in annotation_vars:
            pvals[var] = self.model_fits[var].f_pvalue
        return pvals
    
    def get_fstats_all(self):
        """
        Get the F statistics for all annotation variables
        Returns:
            fstats (dict): dictionary of F statistics
        """
        return self.get_fstats(self.annotation_vars)
    
    def get_pvals_all(self):
        """
        Get the p-values for all annotation variables
        Returns:
            pvals (dict): dictionary of p-values
        """
        return self.get_pvals(self.annotation_vars)

    def get_model_std_errors_all(self):
        """
        Get the model standard errors for all annotation variables
        Returns:
            model_std_errors_all (pandas dataframe): dataframe containing the model standard errors
        """
        model_std_errors_all = {}
        for var in self.annotation_vars:
            model_std_errors_all[var] = self.model_fits[var].bse
        return model_std_errors_all

    def get_model_std_error(self, annotation_var, regressor):
        """
        Get the model standard error for a particular annotation variable and regressor
        Args:
            annotation_var (str): annotation variable
            regressor (str): regressor
        Returns:
            model_std_error (float): model standard error
        """
        return self.model_fits[annotation_var].bse[regressor]

    def get_model_coefficient(self, annotation_var, regressor):
        """
        Get the model coefficient for a particular annotation variable and regressor
        Args:
            annotation_var (str): annotation variable
            regressor (str): regressor
        Returns:
            model_coefficient (float): model coefficient
        """
        return self.model_fits[annotation_var].params[regressor]
    
    def get_model_coefficients_all(self):
        """
        Get the model coefficients for all annotation variables
        Returns:
            model_coefficients_all (pandas dataframe): dataframe containing the model coefficients
        """
        model_coefficients_all = {}
        for var in self.annotation_vars:
            model_coefficients_all[var] = self.model_fits[var].params
        return model_coefficients_all
    
    def get_results(self):
        """
        Get the results for all annotation variables

        Returns:
            results (pandas dataframe): dataframe containing the results
        """
        results = []
        fstats = self.get_fstats_all()
        pvals  = self.get_pvals_all()
        for var in self.annotation_vars:
            results.append(dict(
                annotation=var,
                fstat=fstats[var],
                log_fstat=np.log(fstats[var]),
                pval=pvals[var],
                ))
        return pd.DataFrame(results)


class Efficiency:
    """
    Class to compute the efficiency of a model
    """

    def __init__(self,
                 train_df,
                 test_df,
                 machine_df,
                 annotation_vars,
                 uid_var='uid',
                 act_suffix='_act',
                 pred_suffix='_pred'):
        """
        Args:
            train_df (pandas dataframe): dataframe containing the training data
            test_df (pandas dataframe): dataframe containing the `test data` OR `cross-validation predictions`
            machine_df (pandas dataframe): dataframe containing the unannotated data
            annotation_vars (list): list of annotation variables
            uid_var (str): user id variable
            act_suffix (str): suffix for the actual values
            pred_suffix (str): suffix for the predicted values

        """
        self.train_df = train_df
        self.test_df = test_df
        self.human_df = pd.concat([train_df, test_df])

        self.machine_df = machine_df

        self.annotation_vars = annotation_vars
        self.act_suffix = act_suffix
        self.pred_suffix = pred_suffix

        self.n_m = self.machine_df[uid_var].nunique()
        self.n_h = self.human_df[uid_var].nunique()
        self.n_sum = self.n_h + self.n_m

    def get_variance_machine(self, var):
        """
        Get the variance of the machine data
        Returns:
            sig2_y (float): variance of the machine data
        """
        return self.machine_df[var+self.pred_suffix].var()

    def get_variance_human(self, var):
        """
        Get the variance of the Human data
        Returns:
            sig2_h (float): variance of the Human data
        """
        return self.human_df[var+self.act_suffix].var()

    def get_variance_residual(self, var):
        """
        Get the variance of the residuals
        Returns:
            sig2_eps (float): variance of the residuals
        """
        return (self.test_df[var+self.act_suffix] - self.test_df[var+self.pred_suffix]).var()

    def get_mean_human(self, var):
        """
        Get the mean of the Human data
        Returns:
            mu_h (float): mean of the Human data
        """
        return self.human_df[var+self.act_suffix].mean()

    def get_mean_machine(self, var):
        """
        Get the mean of the Machine data
        Returns:
            mu_m (float): mean of the Machine data
        """
        return self.machine_df[var+self.pred_suffix].mean()

    def get_std_error_human(self, sig2_h):
        """
        Get the standard error of the Human data
        Returns:
            se_h (float): standard error of the annotated data
        """
        return np.sqrt(sig2_h/self.n_h)

    def get_std_error_machine(self, sig2_m):
        """
        Get the standard error of the Machine data
        Returns:
            se_m (float): standard error of the Machine data
        """
        return np.sqrt(sig2_m/self.n_m)

    def get_std_error_enhanced(self, sig2_h, sig2_m):
        """
        Get the standard error of the Enhanced Data
        Returns:
            se_enh (float): standard error of the Enhanced Data
        """
        return np.sqrt(((self.n_h * sig2_h) + (self.n_m * sig2_m))/(self.n_sum**2))

    def get_results(self, var):
        """
        Get the results for a particular annotation variable
        """
        mu_h = self.get_mean_human(var)
        mu_m = self.get_mean_machine(var)

        sig2_h = self.get_variance_human(var)
        sig2_y = self.get_variance_machine(var)
        sig2_eps = self.get_variance_residual(var)

        sig2_m = sig2_y + sig2_eps

        se_h = self.get_std_error_human(sig2_h)
        se_m = self.get_std_error_machine(sig2_m)
        se_enh = self.get_std_error_enhanced(sig2_h, sig2_m)

        return dict(
            annotation=var,
            sig2_m=sig2_m,
            sig2_y=sig2_y,
            sig2_h=sig2_h,
            sig2_eps=sig2_eps,
            mu_h=mu_h,
            mu_m=mu_m,
            se_h=se_h,
            se_m=se_m,
            se_enh=se_enh,
        )

    def get_results_all(self):
        """
        Get the results for all annotation variables
        """
        results = []
        for var in self.annotation_vars:
            results.append(self.get_results(var))
        return pd.DataFrame(results)


# !/usr/bin/python3
"""Feature Scaling & Dimensionality Reduction Classes"""

import umap
import sklearn.cluster
import sklearn.preprocessing
import sklearn.decomposition
from sklearn.base import BaseEstimator, TransformerMixin


class FeatureScaler(BaseEstimator, TransformerMixin):
    """
    Feature Scaling Methods
    """

    def __init__(self, name="StandardScaler", **kwargs):
        """
        Feature Scaling Methods

        Args:
            name (str, optional): Name of the Scaler. Defaults to 'StandardScaler'.
            **kwargs: Keyword arguments to pass to the Scaler.
        """
        self.name = name
        if isinstance(self.name, str):
            self.scaler = getattr(sklearn.preprocessing, self.name)()
            self.scaler.set_params(**kwargs)
        else:
            self.scaler = None

    def get_params(self, **kwargs):
        if self.scaler is not None:
            params = self.scaler.get_params(**kwargs)
            params["name"] = self.name
            return params
        else:
            return {"name": None}

    def set_params(self, **kwargs):
        self.name = kwargs.get("name", "StandardScaler")
        if isinstance(self.name, str):
            self.scaler = getattr(sklearn.preprocessing, self.name)()
            self.scaler.set_params(
                **{k: v for k, v in kwargs.items() if k not in ["name"]}
            )
        else:
            self.scaler = None

    def fit(self, *args, **kwargs):
        """
        Fit the scaler to the data.
        """
        if self.scaler is not None:
            self.scaler.fit(*args, **kwargs)
            return self
        else:
            return self

    def transform(self, X, **kwargs):
        """
        Transform the data.
        """
        if self.scaler is None:
            return X
        else:
            return self.scaler.transform(X, **kwargs)

    def fit_transform(self, X, y=None, **kwargs):
        """
        Fit the scaler to the data and transform it.
        """
        if self.scaler is not None:
            return self.scaler.fit_transform(X, **kwargs)
        else:
            return X


def apply_feature_scaling(X, name="StandardScaler", **kwargs):
    """
    Callable function to apply a `stateless` scaler to a matrix. For use with FunctionTransformer.
    """
    return FeatureScaler(name=name, **kwargs).fit_transform(X)


class DimensionalityReduction(BaseEstimator, TransformerMixin):
    """
    Dimensionality Reduction Methods
    """

    def __init__(self, name="TruncatedSVD", **kwargs):
        """
        Dimensionality Reduction Methods
        Args:
            name (str, optional): Name of the Dimensionality Reduction Method. Defaults to 'TruncatedSVD'.
            **kwargs: Keyword arguments to pass to the Dimensionality Reduction Method.
        """
        self.name = name
        if self.name == "UMAP":
            self.method = umap.UMAP()
        else:
            self.method = getattr(sklearn.decomposition, self.name)()
        self.method.set_params(**kwargs)

    def get_params(self, **kwargs):
        return self.method.get_params(**kwargs)

    def set_params(self, **kwargs):
        self.method.set_params(**kwargs)

    def fit(self, *args, **kwargs):
        """
        Fit the Dimensionality Reduction Method to the data.
        """
        self.method.fit(*args, **kwargs)
        return self

    def transform(self, X, **kwargs):
        """
        Transform the data.
        """
        return self.method.transform(X, **kwargs)

    def fit_transform(self, X, y=None, **kwargs):
        """
        Fit the Dimensionality Reduction Method to the data and transform it.
        """
        return self.method.fit_transform(X, **kwargs)



# /usr/bin/env python
"""Utility functions"""

import pickle
import hashlib
import functools
import numpy as np
import pandas as pd
import simplejson as json
from configparser import NoOptionError
from sklearn.model_selection import train_test_split


def md5_sha_from_str(val: str) -> str:
    """
    Generate md5 and sha256 hash from string.
    """
    return hashlib.md5(val.encode("utf-8")).hexdigest()


def md5_sha_from_dict(obj, ignore_nan=False, default=None):
    """
    Generate md5 and sha256 hash from dictionary.
    """
    json_data = json.dumps(obj, sort_keys=True, ignore_nan=ignore_nan, default=default)
    return md5_sha_from_str(json_data)


def prefix_dict_keys(d, s):
    """
    Prefixes the keys of a dictionary with a string
    """
    return {s + k: v for k, v in d.items()}


def split_dataframe(
    dataframe, key="UID", keepcols=None, train_size=0.75, random_state=None
):
    """
    Split UIDs into train and test sets.
    Use the splits to get train and test dataframes.

    Args:
        dataframe (pd.DataFrame): dataframe to split
        key (str): column name to split on [default: UID]
        keepcols (list): list of columns to keep in train and test dataframes [default: None]
        train_size (float): proportion of data to use for training [default: 0.75]
        random_state (int): random seed [default: None]

    Returns:
        train_df (pd.DataFrame): train dataframe
        test_df (pd.DataFrame): test dataframe
    """
    if train_size == 1.0:
        return dataframe, None
    else:
        unique_keys = dataframe[key].unique().tolist()
        if train_size < 1.0:
            train_keys, test_keys = train_test_split(
                unique_keys, train_size=train_size, random_state=random_state
            )
        elif train_size > 1.0:
            train_keys, test_keys = train_test_split(
                unique_keys, train_size=int(train_size), random_state=random_state
            )
        train_df = dataframe[dataframe[key].isin(train_keys)].reset_index(drop=True)
        test_df = dataframe[dataframe[key].isin(test_keys)].reset_index(drop=True)
        if keepcols is not None:
            train_df = train_df[keepcols]
            test_df = test_df[keepcols]
        return train_df, test_df


def pad_left(x, n=3):
    """
    Converts strings/ integers to 3-digit string code.
    """
    x = str(int(x))
    while len(x) < n:
        x = "0" + x
    return x


def combine_dataframes(dataframes):
    """
    Combine multiple dataframes into one.
    """
    data = pd.concat(dataframes, ignore_index=True)
    return data


def get_probas_from_decisions(X):
    """
    Reverse decision function values to get \
        pseudo-probabilities between 0 & 1

    Args:
        X: decision function output

    Returns: pseudo-probabilities between 0 & 1
        dtype: float
    """
    return 1 / (1 + np.exp(-X))


def save_pickle_data(doc, pkl_filepath):
    """
    Save pickle data to file.
    """
    with open(pkl_filepath, "wb") as pkl_file:
        pickle.dump(doc, pkl_file, protocol=pickle.HIGHEST_PROTOCOL)
    print(f"Saved file in {pkl_filepath}")


def load_pickle_data(pkl_filepath):
    """
    Load pickle data from file.
    """
    with open(pkl_filepath, "rb") as pkl_file:
        doc = pickle.load(pkl_file)
    return doc


def save_json_data(doc, json_filepath):
    """
    Save json data to file.
    """
    with open(json_filepath, "w") as json_file:
        json.dump(doc, json_file)
    print(f"Saved file in {json_filepath}")


def load_json_data(json_filepath):
    """
    Load json data from file.
    """
    with open(json_filepath, "r") as json_file:
        doc = json.load(json_file)
    return doc


def probability_atleast_1(probabilities):
    """
    Calculate the probability of atleast 1 success from a list of probabilities of successes
    """
    return 1 - np.prod([1 - p for p in probabilities])


def uid_aggregate(
    data,
    variables,
    init_suffix="",
    uid_column="uid",
    post_suffix="",
    agg_function="mean",
):
    """
    Aggregate text level records to UID-level records.

    Args:
        data (pd.DataFrame): Dataframe with text level records
        variables (list): List of variables to aggregate
        init_suffix (str): Suffix to add to variables before aggregation
        uid_column (str): UID column name
        post_suffix (str): Suffix to add to variables after aggregation
        agg_function (str): Aggregation function to use

    Returns:
        pd.DataFrame: Dataframe with UID level records
    """
    if isinstance(agg_function, str):
        try:
            agg_function = getattr(np, agg_function)
        except Exception as e:
            raise NoOptionError(
                "`agg_function` must be a numpy function or a custom callable function"
            )
    init_variables = [a + init_suffix for a in variables]
    post_variables = [a + post_suffix for a in variables]
    rename_variables = {
        init_variables[x]: post_variables[x] for x in range(len(variables))
    }

    data_grouped = data[[uid_column, *init_variables]].groupby(uid_column)
    agg_groups = []
    for g, group in data_grouped:
        agg_grp = {uid_column: g}
        agg_grp.update(dict(group[init_variables].apply(agg_function, axis=0)))
        agg_groups.append(agg_grp)
    data_uid = pd.DataFrame(agg_groups)
    data_uid = data_uid.rename(columns=rename_variables)
    return data_uid


def merge_dataframes(dataframes, on="uid", how="inner"):
    """
    Merge multiple dataframes into one.
    """
    return functools.reduce(
        lambda left, right: pd.merge(left, right, on=on, how=how), dataframes
    )


# /usr/bin/env python
"""This script contains common metrics for measuring the classification performance."""

import sklearn.metrics

def count_outcomes(y_true, y_pred, outcome_type="true_positives"):
    """
    Count the number of -
        true positives, false positives, false negatives, true negatives.
    Args:
        y_true: true labels
        y_pred: predicted labels
        outcome_type: "TP", "FP", "FN", or "TN"
    """
    outcomes = [
        "true_negatives",
        "false_positives",
        "false_negatives",
        "true_positives",
    ]
    outcome_values = sklearn.metrics.confusion_matrix(y_true, y_pred).ravel()
    return outcome_values[outcomes.index(outcome_type.lower())]


def calc_f1_score_from_outcomes(
    true_positives, false_positives, false_negatives, true_negatives=None
):
    """
    Calculates the F1 score from a
        True-Positives, False-Postives, and False-Negatives.
    """
    return (2 * true_positives) / (
        2 * true_positives + false_positives + false_negatives
    )


def calc_f1_score_from_labels(y_true, y_pred):
    """
    Calculates the F1 score from true-labels and predicted-labels.
    """
    return sklearn.metrics.f1_score(y_true, y_pred)


def get_scoring_dict(
    keys=[
        "count_true_negatives",
        "count_false_positives",
        "count_false_negatives",
        "count_true_positives",
        "f1",
    ]
):
    """
    Returns a dictionary of scoring functions. Useful for passing to `sklearn.model_selection.cross_validate`.

    Any scoring function in `sklearn.metrics.SCORERS` can be used.

    Args:
        keys: list of scoring functions to return.

    """
    return {key: ALL_SCORERS[key] for key in keys}


def list_scoring_functions():
    """
    List all available scoring functions.
    """
    return list(ALL_SCORERS.keys())


SKLEARN_SCORERS = sklearn.metrics.SCORERS.copy()
MORE_SCORERS = {
    "count_"
    + outcome: sklearn.metrics.make_scorer(
        count_outcomes, needs_proba=False, needs_threshold=False, outcome_type=outcome
    )
    for outcome in [
        "true_negatives",
        "false_positives",
        "false_negatives",
        "true_positives",
    ]
}
ALL_SCORERS = {**SKLEARN_SCORERS, **MORE_SCORERS}


import re
import unicodedata
import numpy as np
import num2words

# List all functions
__all__ = [
    "remove_accents",
    "remove_extra_whitespaces",
    "remove_extra_newlines",
    "remove_extra_tabs",
    "remove_extra_spaces",
    "split_after_colon",
    "lowercase",
    "remove_quotes",
    "remove_special_chars",
    "remove_space_between_numbers",
    "remove_dashes",
    "remove_commas_between_numbers",
    "replace_numbers",
    "keep_alnum_only",
    "keep_whitespace_alnum",
    "count_words",
    "replace_contractions",
    "mask_digits",
    "remove_stopwords",
    "remove_short_words",
    "replace_whole_words",
]


def remove_accents(text):
    """"""
    nfkd_form = unicodedata.normalize("NFKD", text)
    return "".join([c for c in nfkd_form if (unicodedata.combining(c) == 0)])


def remove_extra_whitespaces(s):
    """
    Removes extra spaces from strings
    """
    return re.sub(" +", " ", s).strip()


def remove_extra_newlines(s):
    """
    Removes extra newlines from strings
    """
    return re.sub("\n+", "\n", s)


def remove_extra_tabs(s):
    """
    Removes extra tabs from strings
    """
    return re.sub("\t+", "\t", s)


def remove_extra_spaces(s):
    """
    Removes extra spaces from strings
    """
    s = remove_extra_whitespaces(s)
    s = remove_extra_newlines(s)
    s = remove_extra_tabs(s)
    return s


def split_after_colon(text, max_left=10):
    """
    Splits text after every colon.
    """
    splits = re.split("ঃ|:", text)
    if len(splits) == 2 and len(splits[0]) <= max_left:
        text = splits[-1]
    return text.strip()


def lowercase(text):
    """Converts all characters to lowercase"""
    if isinstance(text, str):
        return text.lower()
    else:
        return text


def remove_quotes(string):
    """
    Removes all quotes Single(') or Double(") from strings
    """
    pattern = re.compile("'|\"")
    return re.sub(pattern, "", string)


def remove_special_chars(s):
    """
    Removes the following special characters from strings
    """
    s = re.sub("[\\\/!,*)@#%(&$_.?^-]", " ", s)
    s = remove_quotes(s)
    return re.sub(" +", " ", s).strip()


def remove_space_between_numbers(text):
    """
    Removes whitespaces between numbers
    """
    text = re.sub(r"(\d)\s+(\d)", r"\1\2", text)
    return text


def remove_dashes(x):
    """
    Removes dashes from strings
    """
    x = re.sub("-+", " ", x)
    x = re.sub(" +", " ", x)
    return x


def remove_commas_between_numbers(text):
    """
    Removes commas between numbers
    """
    text = re.sub(r"(\d)\s*,\s*(\d)", r"\1\2", text)
    return text


def replace_numbers(text):
    """
    Replaces all numbers with a given string
    """
    return re.sub(r"(\d+)", lambda x: num2words.num2words(int(x.group(0))), text)


def keep_alnum_only(text):
    """
    Removes all non-alphanumeric characters from strings
    NOTE: This function will remove whitespaces as well.
    """
    return "".join([a for a in text if a.isalnum()])


def keep_whitespace_alnum(text):
    """
    Removes all non-alphanumeric characters from strings
    Keeps whitespaces, and alpha-numeric characters
    """
    if isinstance(text, str):
        text = "".join([a if a.isalnum() else " " for a in text])
        text = remove_extra_spaces(text)
        return text
    else:
        return text


def count_words(text):
    """
    Counts the number of words in a string
    """
    text = keep_whitespace_alnum(text)
    return len([w for w in text.split() if w.isalnum()])


def mask_digits(s):
    """
    Experimental function to mask digits in strings
    Replaces all digits with '#'
    """
    s = "".join(["#" if x.isdigit() else x for x in s])
    digit_counts = list(
        np.unique([(i.end() - i.start()) for i in list(re.finditer("#+", s))])
    )
    digit_counts.sort()
    digit_counts.reverse()
    for d in digit_counts:
        s = re.sub("".join(["#" for _ in range(d)]), f" {d}# ", s)
    return re.sub(" +", " ", s).strip()


def remove_stopwords(text: str, stopwords):
    """
    Removes all stopwords from strings
    """
    return " ".join([t for t in text.split() if t not in stopwords])


def remove_short_words(text: str, min_length=2):
    """
    Removes all words with length less than min_length
    """
    return " ".join([t for t in text.split() if len(t) >= min_length])


def replace_whole_words(text, replace_words: dict):
    """
    Replaces certain words with a given string
    For example:
        'Yes' is often incorrectly tagged as 'G' in the dataset
    So we replace all 'G'/'G.' with 'Yes'
    """
    return " ".join(
        [replace_words[t] if t in replace_words else t for t in text.split()]
    )



"""Post-Vectorization Estimation Module"""

import sklearn
import warnings
import numpy as np
import sklearn.svm
import sklearn.isotonic
import sklearn.linear_model
import sklearn.semi_supervised
import sklearn.neural_network
import sklearn.naive_bayes
import sklearn.tree
import sklearn.neighbors
import sklearn.ensemble
from sklearn.base import BaseEstimator, TransformerMixin
from .evaluation import EvaluationScorer

module_options = {
    "naive_bayes": sklearn.naive_bayes.__all__,
    "linear_model": sklearn.linear_model.__all__,
    "semi_supervised": sklearn.semi_supervised.__all__,
    "svm": sklearn.svm.__all__,
    "isotonic": sklearn.isotonic.__all__,
    "tree": sklearn.tree.__all__,
    "ensemble": sklearn.ensemble.__all__,
    "neighbors": sklearn.neighbors.__all__,
    "neural_network": sklearn.neural_network.__all__,
}
module_dict_list = [
    {f: module for f in funcmethods} for module, funcmethods in module_options.items()
]
estimator_options = {k: v for element in module_dict_list for k, v in element.items()}


def get_probas_from_decisions(X):
    """
    Reverse decision function values to get \
        pseudo-probabilities between 0 & 1

    Args:
        X: decision function output

    Returns: pseudo-probabilities between 0 & 1
        dtype: float
    """
    return 1 / (1 + np.exp(-X))


def load_estimator(model, estimator_options=estimator_options):
    """
    Load an estimator from sklearn.
    """
    if model in estimator_options:
        return getattr(getattr(sklearn, estimator_options.get(model)), model)
    else:
        raise ValueError("{} is not a valid estimator model name".format(model))


class BinaryThresholder(BaseEstimator):
    """
    Custom Threshold Predictor to be used in the pipeline.
    """

    def __init__(
        self,
        use_threshold=True,
        threshold_range=(0.001, 0.999),
        steps=100,
        threshold=0.5,
        scoring_metric="f1_score",
    ):
        """
        Args:
            use_threshold (bool): Whether to use thresholding or not. Defaults to True.
            threshold_range (tuple): Range of thresholds to search over. Defaults to (0.001, 0.999).
            steps (int): Number of steps to search over. Defaults to 100.
            threshold (float): Threshold to use. Defaults to 0.5.
            scoring_metric (str): Scoring metric to use. Defaults to "f1_score".
        """
        self.use_threshold = use_threshold
        self.steps = steps
        self.threshold_range = threshold_range
        self.thresholds = np.linspace(
            self.threshold_range[0], self.threshold_range[1], self.steps
        )
        self.scoring_metric = scoring_metric
        self.scorer = EvaluationScorer(metrics=[self.scoring_metric])
        self.threshold = threshold
        self.params = {
            "use_threshold": self.use_threshold,
            "threshold_range": self.threshold_range,
            "steps": self.steps,
            "threshold": self.threshold,
            "scoring_metric": self.scoring_metric,
        }

    def get_params(self, **kwargs):
        return self.params.copy()

    def set_params(self, **kwargs):
        """ """
        self.use_threshold = kwargs.get("use_threshold", self.use_threshold)
        self.threshold_range = kwargs.get("threshold_range", self.threshold_range)
        self.steps = kwargs.get("steps", self.steps)
        self.thresholds = np.linspace(
            self.threshold_range[0], self.threshold_range[1], self.steps
        )
        self.scoring_metric = kwargs.get("scoring_metric", self.scoring_metric)
        self.scorer = EvaluationScorer(metrics=[self.scoring_metric])
        self.threshold = kwargs.get("threshold", self.threshold)
        self.params = {
            "use_threshold": self.use_threshold,
            "threshold_range": self.threshold_range,
            "steps": self.steps,
            "threshold": self.threshold,
            "scoring_metric": self.scoring_metric,
        }

    def fit(self, X=None, Y=None, *args, **kwargs):
        """
        Fit the estimator.
        Args:
            X : probas of the classifier
            Y : labels
        Returns:
            self
        """
        if self.use_threshold is True:
            t0, t1 = min(X[:, 1]), max(X[:, 1])
            self.threshold_range = (t0, t1)
            self.thresholds = np.linspace(t0, t1, self.steps)
            y_pred_array = [self.predict(X, thresh=t) for t in self.thresholds]
            scores = [
                self.scorer.calc_scores(Y, y_preds)[self.scoring_metric].values[0]
                for y_preds in y_pred_array
            ]
            self.threshold = self.thresholds[np.argmax(scores)]
        else:
            self.threshold = None
            self.thresholds = []
            self.steps = 0
            self.threshold_range = None
        self.set_params(threshold=self.threshold, use_threshold=self.use_threshold)
        return self

    def predict_proba(self, X, **kwargs):
        """
        Predict probabilities of the positive class.
        """
        return X

    def predict_log_proba(self, X, **kwargs):
        """
        Predict the log probability of the positive class.
        """
        return np.log(X)

    def decision_function(self, X, **kwargs):
        """ """
        log_probs = self.predict_log_proba(X)
        return log_probs[:, 1] - log_probs[:, 0]

    def predict(self, X, thresh=None, validate=False, **pred_kwargs):
        """
        Predict the class labels for the provided data.
        """
        if self.use_threshold is False:
            return X
        else:
            if thresh is not None:
                t = thresh
            else:
                t = self.threshold
            if validate is False:
                return [1 if v >= t else 0 for v in X[:, 1]]
            else:
                proba_valid = self.validate_proba(X[:, 1])
                if proba_valid is True:
                    return [1 if v >= t else 0 for v in X[:, 1]]
                else:
                    return [None] * len(X[:, 1])

    def validate_proba(self, probas, **kwargs):
        """
        Check if the probabilities are valid
        """
        if len(set(probas)) > 1:
            return True
        else:
            return False


class Classifier(BaseEstimator, TransformerMixin):
    """
    Main classifier methods
    """

    def __init__(self, model, is_final=False, **kwargs) -> callable:
        """ """
        self.model = model
        self.is_final = is_final
        self.classifier = load_estimator(model=self.model)(**kwargs)

    def get_params(self, **kwargs) -> dict:
        params = self.classifier.get_params(**kwargs)
        params["model"] = self.model
        params["is_final"] = self.is_final
        return params

    def set_params(self, **kwargs):
        self.model = kwargs.get("model", "LogisticRegression")
        self.is_final = kwargs.get("is_final", False)
        self.classifier = load_estimator(model=str(self.model))()
        self.classifier.set_params(
            **{k: v for k, v in kwargs.items() if k not in ["model", "is_final"]}
        )

    def fit(self, *args, **kwargs):
        """
        Fit the estimator.
        """
        self.classifier.fit(*args, **kwargs)
        return self

    def get_normalized_decisions(self, X, **kwargs):
        """
        Artificial probabilities - normalized decisions.
            There's no need to fit the estimator for training,
                since we will only use it on the decision-values to classify.
        """
        if hasattr(self.classifier, "decision_function"):
            decisions = self.classifier.decision_function(X, **kwargs)
        elif hasattr(self.classifier, "_decision_function"):
            decisions = self.classifier._decision_function(X, **kwargs)
        else:
            print("No decision function found.")
        positive_proba = get_probas_from_decisions(decisions)
        negative_proba = 1 - positive_proba
        return np.vstack((negative_proba, positive_proba)).T

    def get_coefficient(self, X, **kwargs):
        """ """
        if hasattr(self.classifier, "coef_"):
            return self.classifier.coef_
        else:
            warnings.warn("No coefficient found.")
            return None

    def transform(self, X, **kwargs):
        """ """
        return self.predict_proba(X, **kwargs)

    def predict_proba(self, X, **kwargs):
        """
        Predict the probability of each class for each sample.
        """
        # If `predict_proba` is available, use it
        if hasattr(self.classifier, "predict_proba"):
            return self.classifier.predict_proba(X, **kwargs)
        else:
            # Otherwise, use `decision_function`
            if hasattr(self.classifier, "decision_function") or hasattr(
                self.classifier, "_decision_function"
            ):
                return self.get_normalized_decisions(X, **kwargs)
            else:
                warnings.warn("No probability function found. Returning decisions.")
                return self.predict(X, **kwargs)

    def predict(self, X, **kwargs):
        """
        Predict the class for each sample.
            The last layer of the pipeline will use this function.
            As long as the thresholdlayer is called, this function will not be used.
        """
        predictions = self.classifier.predict(X, **kwargs)
        return np.nan_to_num(predictions, nan=0)



from setuptools import setup, find_packages

# Read requirements.txt
with open('requirements.txt') as f:
    required = f.read().splitlines()

# Read README.md
with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

# Setup
setup(
    name='iQual',
    version='0.1.0',
    description='iQual is a package that leverages natural language processing to scale up interpretative qualitative analysis. It also provides methods to assess the bias, interpretability and efficiency of the machine-enhanced codes.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Aditya Karan Chhabra',
    author_email='aditya0chhabra@gmail.com',
    url='https://github.com/worldbank/iQual',
    maintainer='Aditya Karan Chhabra',
    maintainer_email='aditya0chhabra@gmail.com',
    packages=find_packages(where='src'),
    exclude_package_data={'': ['data','notebooks','docs','tests']},
    license='MIT',
    license_files=('LICENSE.md',),
    package_dir={'': 'src'},
    install_requires=required,
    python_requires='>=3.7',
    keywords='nlp natural-language-processing qualitative-analysis human-coding qualitative-research',
    classifiers=[
        'Development Status :: 3 - Alpha',
        "Intended Audience :: Science/Research",
        "Topic :: Sociology",
        "Topic :: Utilities",
        "Topic :: Text Processing :: Linguistic",
        "Topic :: Scientific/Engineering :: Information Analysis",
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
)



# /usr/bin/env python
"""This script contains cross-validation methods and classes used by the iQual package"""

import operator
import warnings
import functools
import pandas as pd
import numpy as np
import sklearn.model_selection
from collections import Counter


def convert_dict(d, prefix):
    """
    Convert a nested dictionary to a flat dictionary
    """

    result = {}
    for k, v in d.items():
        if isinstance(v, dict):
            prefix_fmt = f"{prefix}__{k}".strip("__")
            result.update(convert_dict(v, prefix_fmt))
        elif isinstance(v, list):
            prefix_fmt = f"{prefix}__{k}".strip("__")
            result[prefix_fmt] = v
    return result


def convert_nested_params(configuration):
    """
    Convert a nested CV configuration to a flat dictionary
    """

    return convert_dict(configuration, "")


def count_hyperparameters(param_grid):
    """
    Count hyperparameter grid of type list-or-dict
    """

    def count_dict_params(param_dict):
        """
        Count hyperparameter grid of type dict
        """
        num_configurations = functools.reduce(
            operator.mul, [len(p) for p in param_dict.values()], 1
        )
        return num_configurations

    if isinstance(param_grid, dict):
        num_params = count_dict_params(param_grid)
    elif isinstance(param_grid, list):
        num_params = sum([count_dict_params(pdict) for pdict in param_grid])
    else:
        num_params = None
    return num_params


class CrossValidator:
    """
    Cross-validation methods
    """

    def __init__(
        self,
        model_pipe,
        search_parameters,
        cv_method="GridSearchCV",
        scoring="f1",
        cv=3,
        refit=False,
        **kwargs,
    ):
        """
        Needs a valid `cv_method` from `sklearn.model_selection`
        Configure the cross-validation object
        """

        self.cross_validation = getattr(sklearn.model_selection, cv_method)
        self.cross_validation = self.cross_validation(
            model_pipe,
            search_parameters,
            scoring=scoring,
            cv=cv,
            refit=refit,
            **kwargs,
        )

    def fit(self, X, Y, **kwargs):
        """
        Fit the cross-validation object
        """

        self.cross_validation.fit(X, Y, **kwargs)
        self.refit_string = self.cross_validation.refit
        self.n_splits = self.cross_validation.n_splits_
        if not self.refit_string:
            self.refit_string = "score"
        self.rank_column = "rank_test_" + self.refit_string
        self.best_parameters = self.get_best_params()
        self.param_df = self.get_cv_results()

        self.cv_scores = self.param_df.loc[
            self.param_df[self.rank_column] == 1,
            [f"split{i}_test_{self.refit_string}" for i in range(self.n_splits)],
        ].to_dict("records")[0]
        avg_cv_score = np.mean(list(self.cv_scores.values()))
        self.cv_scores.update({"avg_test_score": avg_cv_score})

    def get_cv_scores(self):
        """
        Get the cross-validation scores
        """
        return self.cv_scores.copy()

    def get_cv_results(self):
        """
        Get the cross-validation results as a dataframe
        """
        self.param_df = pd.DataFrame(self.cross_validation.cv_results_)
        self.param_df = self.param_df.sort_values(
            by=self.rank_column, ignore_index=True
        )
        return self.param_df.copy()

    def predict(self, X):
        """
        Predict using the best estimator
        """
        return self.cross_validation.predict(X)

    def get_best_params(self):
        """
        Get the best parameters
        """
        return self.cross_validation.best_params_


class Splitter:
    """
    Customized CV Splitter class.

    To ensure that a minimum threshold of positive class is satisfied for the split

    """

    def __init__(self, method: str, thresh=0.2, attempts=3, **kwargs) -> None:
        """
        Initialize the splitter object.
        """
        self.method = method
        self.thresh = thresh
        self.valid_split = True
        self.attempts = attempts
        self.kwargs = kwargs
        self.splitter = getattr(sklearn.model_selection, method)(**kwargs)

    def get_n_splits(self, X, y=None, groups=None):
        """
        Returns the number of splitting iterations in the cross-validator

        """
        return self.splitter.get_n_splits(X, y, groups)

    def split_and_verify(self, X, y=None, groups=None) -> list:
        """
            NOTE: Threshold: 20% of total positives

        Split the data and verify that the minimum threshold of positive class is satisfied

        split_list: list of tuples of (train_index, test_index)
        """
        total_positives = Counter(y)[1]
        n_splits = self.splitter.get_n_splits(X, y, groups=groups)
        if total_positives < n_splits:
            self.valid_split = False
        else:
            self.valid_split = True
            split_generator = self.splitter.split(X, y, groups=groups)
            split_list = [next(split_generator) for _ in range(n_splits)]
            for n in range(n_splits):
                y_values = [y[j] for j in split_list[n][1]]
                y_count = Counter(y_values)[1]
                pct_thresh = y_count / total_positives
                if pct_thresh > self.thresh:
                    continue
                else:
                    self.valid_split = False
                    break
            return iter(split_list)

    def split(self, X, y=None, groups=None):
        """
        Generate indices to split data into training and test set.
        """
        for _ in range(self.attempts):
            split_generator = self.split_and_verify(X, y, groups)
            if self.valid_split:
                break
            else:
                continue
        if not self.valid_split:
            warnings.warn("Splitter Threshold not met")
        return split_generator

    def get_datasets(self, X, y=None):
        """
        Get the datasets for each split
        """
        split_list = self.split(X, y)
        datasets = []
        for train_index, test_index in split_list:
            X_train, X_test = X.loc[train_index], X.loc[test_index]
            y_train, y_test = y[train_index], y[test_index]
            datasets.append(
                {
                    "train_X": X_train,
                    "test_X": X_test,
                    "train_y": y_train,
                    "test_y": y_test,
                }
            )
        return datasets
